import matplotlib.patches as mpatches


class ArrowDrawer:
    def plot_arrows(self, arrowstyle="->"):
        if self.arrows and len(self.arrows) > 0:
            for arr in self.arrows:
                arr_patch = mpatches.FancyArrowPatch(arr[0], arr[1], arrowstyle=arrowstyle, mutation_scale=10)
                self.ax.add_patch(arr_patch)


class ArrowDrawerTernary:
    def plot_arrows(self, arrowstyle="->"):  # Plot arrow with normal x vs y projection
        if self.arrows and len(self.arrows) > 0:

            for arr in self.arrows:
                arr_patch = mpatches.FancyArrowPatch(arr[0], arr[1], arrowstyle=arrowstyle, mutation_scale=10)
                self.ax.add_patch(arr_patch)


class LegendDrawer:
    def plot_legend(self):
        if not self.no_legend:
            leg_canvas = self.ax if self.legend_in_axs else self.fig
            leg = leg_canvas.legend(loc=self.legend_loc, ncol=self.legend_ncol, fontsize=self.legend_fs, edgecolor='k')

            # Set marker size in legend
            for leg_obj in leg.legendHandles:
                leg_obj._sizes = self.legend_ms
        else:
            self.fig.subplots_adjust(bottom=0.11)



import random
from matplotlib import colors as mplcol


def darken_color(color, percent=10):
    fcol = list(mplcol.to_rgba(color))
    for i in range(4):
        fcol[i] = max(0, fcol[i] * (1 - percent / 100))
    return fcol


def near_color(color):
    col = mplcol.to_rgba_array(color)[0]

    cr = col[0]
    cg = col[1]
    cb = col[2]

    cr = cr + (-100 + random.randint(0, 200)) / 255.
    cg = cg + (-100 + random.randint(0, 200)) / 255.
    cb = cb + (-100 + random.randint(0, 200)) / 255.

    return (cr, cg, cb, 1)



import warnings

import matplotlib.patches as patches
import matplotlib.pyplot as plt
from matplotlib import lines

from georunes.plot.base import DiagramBase
from georunes.tools.chemistry import val_ox_to_el_ppm, name_el_to_def_ox
from georunes.tools.data import row_min, row_max
from georunes.tools.preprocessing import check_geochem_res
from georunes.tools.reservoirs import Reservoirs

listing_ree = ["La", "Ce", "Pr", "Nd", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu"]


class DiagramSpider(DiagramBase):
    def __init__(self, datasource,
                 listing=listing_ree,
                 show_reservoirs=["", ],
                 norm="CI",
                 h_ratio=1. / 2,
                 fillmode="marked-lines",
                 ylim=None,
                 legend_fs="small", thick_legend_linewidth=False,
                 enclosed_in_bg=("",),
                 markersize=8,
                 xlabel="", ylabel="",
                 drawing_order="zorder",
                 **kwargs
                 ):

        if fillmode not in ("lines", "marked-lines", "enclosed", "enclosed-lines", "mixed"):
            raise ValueError("Incorrect parameter 'fillmode'")

        DiagramBase.__init__(self, datasource=datasource, h_ratio=h_ratio, **kwargs)

        self.thick_legend_linewidth = thick_legend_linewidth
        self.listing = listing
        self.show_reservoirs = (*show_reservoirs,)

        res = Reservoirs.get_instance()

        if isinstance(norm, dict) and check_geochem_res(norm):
            self.norm = res.register_model(norm)
        elif norm in res.model_list:
            self.norm = res.compos[norm]
        else:
            if norm is not None:
                msg = "The data of reservoir " + norm + " are not found. Check the figure configuration."
                warnings.warn(msg)
            self.norm = res.compos["CI"]
        self.fillmode = fillmode
        self.enclosed_in_bg = enclosed_in_bg
        self.drawing_order = drawing_order if drawing_order in self.data.columns else None
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.ylim = ylim
        self.markersize = markersize
        self.legend_fs = legend_fs
        if enclosed_in_bg is None:
            self.enclosed_in_bg = ("",)
        else:
            self.enclosed_in_bg = enclosed_in_bg

    def init_plot(self):

        if self.h_ratio is None:
            self.fig, self.ax = plt.subplots()
        else:
            if isinstance(self.h_ratio, list):
                self.fig, self.ax = plt.subplots(figsize=self.h_ratio)
            else:
                self.fig, self.ax = plt.subplots(figsize=plt.figaspect(self.h_ratio))

    def set_decoration(self):

        if not self.no_title:
            self.ax.set_title(self.title, size="small")

        if self.ylim:
            self.ax.set_ylim(self.ylim)

        self.ax.tick_params(axis='x', labelsize=self.fontsize)
        self.ax.tick_params(axis='y', labelsize=self.fontsize)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

    def is_group_enclosed(self, name):
        if type(name) != str and len(
                name) > 1:  # If multiple parameters are chosen to filter group data,
            # take the first one who must be the group
            name = name[0]

        if (self.enclosed_in_bg is not None) and (str(name) in self.enclosed_in_bg):
            return True
        return False

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        label_defined = True if 'label' in self.data.columns else False

        d_min = dict()
        d_max = dict()

        # Draw spectras
        groups = self.data.groupby(self.group_name)
        legend_cfg = []
        legend_cfg_labels = []
        legend_list = []

        for name, group in groups:
            if self.is_group_allowed(name):
                if self.is_group_enclosed(name):
                    custom_fillmode = 'enclosed'
                else:
                    custom_fillmode = self.fillmode

                last = None
                label = list(group['label'])[0] if label_defined else name
                min_vals = [None] * len(self.listing)  # Trick to deal with missing value
                max_vals = [None] * len(self.listing)
                lw = 0.3  # Linewidth

                for index, row in group.iterrows():
                    vals, calc_val = [], []
                    drow = row.to_dict()
                    for elt in self.listing:
                        if elt in drow.keys():
                            calc_val = drow[elt] / self.norm[elt]
                        elif elt in name_el_to_def_ox.keys():
                            ox = name_el_to_def_ox[elt]
                            calc_val = val_ox_to_el_ppm(drow[ox], ox) / self.norm[elt]

                        vals.append(calc_val)
                        if calc_val:
                            if elt in d_min.keys():
                                d_min[elt] = min(d_min[elt], calc_val)
                            else:
                                d_min[elt] = calc_val

                            if elt in d_max.keys():
                                d_max[elt] = max(d_max[elt], calc_val)
                            else:
                                d_max[elt] = calc_val

                    zorder = None
                    if self.drawing_order:
                        zorder = list(group[self.drawing_order])[0]
                    if custom_fillmode in ("enclosed-lines", "mixed", "enclosed"):
                        min_vals = row_min(min_vals, vals)
                        max_vals = row_max(max_vals, vals)
                    mark = list(group["marker"])[0]
                    marker_edge_w = None
                    if mark in ("+", "x"):
                        marker_edge_w = 3

                    if custom_fillmode in ("enclosed-lines", "lines"):
                        last, = self.ax.semilogy(self.listing, vals, markersize=0,
                                                 c=drow['color'],
                                                 mec=drow['color'],
                                                 linewidth=lw,
                                                 alpha=0.8,
                                                 label=label, zorder=zorder
                                                 )

                    elif custom_fillmode in ('marked-lines', 'mixed'):
                        last, = self.ax.semilogy(self.listing,
                                                 vals,
                                                 c=drow['color'],
                                                 mec=drow['color'],
                                                 marker=mark,
                                                 linewidth=lw,
                                                 markeredgewidth=marker_edge_w,
                                                 alpha=0.8,
                                                 label=label,
                                                 markersize=self.markersize
                                                 , zorder=zorder)

                    elif custom_fillmode == "enclosed":  # Fake draw to deal with figure canvas
                        last, = self.ax.semilogy(self.listing, vals,
                                                 linewidth=0, zorder=zorder
                                                 )

                if custom_fillmode in ("enclosed", "enclosed-lines", "mixed"):
                    self.ax.fill_between(self.listing, min_vals, max_vals, facecolor=list(group["color"])[0],
                                         zorder=zorder,
                                         alpha=0.25)

                    # For legend
                    pseudo_square = patches.Rectangle((0, 0), 0, 0, facecolor=list(group["color"])[0], alpha=0.25)
                    pseudo_line = lines.Line2D((0, 0), (0, 0), linewidth=lw * 1.8, c=drow['color'], )

                if name not in legend_list:
                    legend_cfg_labels.append(label)
                    if custom_fillmode == "enclosed":
                        legend_cfg.append(pseudo_square)
                    elif custom_fillmode == "enclosed-lines":
                        legend_cfg.append((pseudo_square, pseudo_line))
                    elif custom_fillmode == "lines":
                        legend_cfg.append(last)
                    elif custom_fillmode == "mixed":
                        legend_cfg.append((pseudo_square, last))
                    elif custom_fillmode == "marked-lines":
                        legend_cfg.append(last)
                    legend_list.append(name)

        self.plot_reservoirs(legend_cfg, legend_list, legend_cfg_labels)
        self.plot_legend(legend_cfg, legend_cfg_labels)

    def plot_reservoirs(self, legend_cfg, legend_list, legend_cfg_labels):
        res = Reservoirs.get_instance()

        for model in self.show_reservoirs:
            if model in res.model_list:
                compo = res.compos[model]

                vals = []
                for val in self.listing:
                    vals.append(compo[val] / self.norm[val])

                last, = self.ax.semilogy(self.listing, vals,
                                         c=res.get_color(model), marker="o", markersize=2,
                                         linewidth=1.5, alpha=0.8,
                                         label=res.get_label(model))
                if model not in legend_list:
                    legend_cfg.append(last)
                    legend_list.append(model)
                    legend_cfg_labels.append(res.get_label(model))

    def plot_legend(self, legend_cfg, legend_cfg_labels):

        if not self.no_legend:

            leg_canvas = self.ax if self.legend_in_axs else self.fig
            leg = leg_canvas.legend(handles=tuple(legend_cfg), labels=tuple(legend_cfg_labels), loc=self.legend_loc,
                                    ncol=self.legend_ncol, edgecolor='k', fontsize=self.legend_fs)
            if self.thick_legend_linewidth:

                for leg_obj in leg.legendHandles:
                    leg_obj.set_linewidth(2.0)

            self.fig.subplots_adjust(bottom=0.2)



import numpy as np
from georunes.plot.base import DiagramBase
from georunes.plot.helpers import ArrowDrawer, LegendDrawer
from georunes.tools.chemistry import val_ox_to_mc


# Debon, F. and Le Fort, P., 1983. A chemical–mineralogical classification of common plutonic rocks and associations.
# Earth and Environmental Science Transactions of the Royal Society of Edinburgh, 73(3), pp.135-149.

class DiagramPQ(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title="P-Q multicationic classification (Debon and Le Fort 1983)",
                 padding={"bottom": 0.20}, annotation=None, **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, padding=padding, **kwargs)

        self.annotation = annotation
        self.xlabel = "P = K - (Na+Ca)"
        self.ylabel = "Q = Si/3 - (K+Na+2Ca/3)"

    def set_decorations(self):

        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        self.ax.set_xlim(-400, 110)
        self.ax.set_ylim(0, 260)

        lx1 = np.array([-377, -257, -205, -224, -311, -377])
        ly1 = np.array([0, 0, 55, 57, 57, 0])
        lx2 = np.array([-257, -188, -143, -205, -257])
        ly2 = np.array([0, 0, 48, 55, 0])
        lx3 = np.array([-188, -117, -88, -143, -188])
        ly3 = np.array([0, 0, 43, 48, 0])
        lx4 = np.array([-117, 16, 30, -40, -88, -117])
        ly4 = np.array([0, 0, 26, 38, 43, 0])
        lx5 = np.array([-311, -224, -205, -187, -172, -274, -311])
        ly5 = np.array([57, 57, 55, 79, 104, 100, 57])
        lx6 = np.array([-205, -143, -125, -103, -172, -187, -205])
        ly6 = np.array([55, 48, 70, 107, 104, 79, 55])
        lx7 = np.array([-143, -88, -69, -55, -103, -125, -143])
        ly7 = np.array([48, 43, 77, 109, 107, 70, 48])
        lx8 = np.array([-88, -40, 30, 64, -30, -55, -69, -88])
        ly8 = np.array([43, 38, 26, 110, 110, 109, 77, 43])
        lx9 = np.array([-274, -172, -156, -134, -119, -216, -230, -256, -274])
        ly9 = np.array([100, 104, 136, 190, 242, 238, 190, 132, 100])
        lx10 = np.array([-172, -103, -89, -71, -59, -119, -134, -156, -172])
        ly10 = np.array([104, 107, 137, 190, 244, 242, 190, 136, 104])
        lx11 = np.array([-103, -55, -44, -29, -17, -59, -71, -89, -103])
        ly11 = np.array([107, 109, 140, 190, 245, 244, 190, 137, 107])
        lx12 = np.array([-55, -30, 64, 75, 90, 105, -17, -29, -44, -55])
        ly12 = np.array([109, 110, 110, 140, 190, 250, 245, 190, 140, 109])

        self.ax.plot([0, 0], [0, 300], color=self.decor_line_col, linewidth=1)

        self.ax.plot(lx1, ly1, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx4, ly4, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx5, ly5, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx6, ly6, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx7, ly7, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx8, ly8, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx9, ly9, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx10, ly10, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx11, ly11, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx12, ly12, color=self.decor_line_col, linewidth=0.6)

        self.ax.text(-170, 230, "to", size='x-small', color=self.decor_text_col)
        self.ax.text(-100, 230, "gd", size='x-small', color=self.decor_text_col)
        self.ax.text(-45, 230, "ad", size='x-small', color=self.decor_text_col)
        self.ax.text(25, 230, "gr", size='x-small', color=self.decor_text_col)
        self.ax.text(-240, 90, "dq", size='x-small', color=self.decor_text_col)
        self.ax.text(-150, 90, "mzdq", size='x-small', color=self.decor_text_col)
        self.ax.text(-75, 90, "mzq", size='x-small', color=self.decor_text_col)
        self.ax.text(-20, 90, "sq", size='x-small', color=self.decor_text_col)
        self.ax.text(-300, 25, "go", size='x-small', color=self.decor_text_col)
        self.ax.text(-215, 25, "mzgo", size='x-small', color=self.decor_text_col)
        self.ax.text(-145, 25, "mz", size='x-small', color=self.decor_text_col)
        self.ax.text(-45, 25, "s", size='x-small', color=self.decor_text_col)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                si = val_ox_to_mc(group["SiO2"])
                na = val_ox_to_mc(group["Na2O"])
                k = val_ox_to_mc(group["K2O"])
                ca = val_ox_to_mc(group["CaO"])

                param_P = k - (na + ca)
                param_Q = si / 3 - (k + na + 2 * ca / 3)

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.ax.scatter(param_P, param_Q, edgecolors=group["color"],
                                marker=list(group["marker"])[0], label=name, facecolors=group["color"],
                                s=self.markersize,
                                alpha=0.9, zorder=zorder)

                if self.annotation:
                    for i, sample in param_P.iteritems():
                        self.ax.annotate(group[self.annotation].get(i), (param_P.get(i), param_Q[i]),
                                         fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



from matplotlib.colors import to_rgba
from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.chemistry import val_el_to_mol
from georunes.tools.plotting import normalize_marker_size


class DiagramVs(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource,
                 xvar, yvar, xlim=None, ylim=None,
                 xlabel='', ylabel='',
                 annotation=None,
                 xscale='linear', yscale='linear',
                 padding=None,
                 marker='',
                 xmolar=False, ymolar=False,
                 alpha_color=0.4, alpha_edge_color=0.8,
                 x_formatter=None, y_formatter=None,
                 markersize=70,
                 **kwargs
                 ):
        config_padding = {"bottom": 0.20}
        if padding:
            config_padding.update(padding)
        DiagramBase.__init__(self, datasource=datasource,
                             title='Diagram ' + xlabel + " vs " + ylabel,
                             padding=config_padding, markersize=markersize,
                             **kwargs)

        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlim = xlim
        self.ylim = ylim
        self.xvar = xvar
        self.yvar = yvar
        self.annotation = annotation
        self.xscale = xscale
        self.yscale = yscale
        self.marker = marker
        self.alpha_color = alpha_color
        self.alpha_edge_color = alpha_edge_color
        self.xmolar = xmolar
        self.ymolar = ymolar
        self.x_formatter = x_formatter
        self.y_formatter = y_formatter

    def set_decorations(self):

        if self.xlim is not None:
            self.ax.set_xlim(self.xlim)
        if self.ylim is not None:
            self.ax.set_ylim(self.ylim)

        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)

        self.ax.set_yscale(self.yscale)
        self.ax.set_xscale(self.xscale)
        self.ax.tick_params(axis='x', labelsize=self.fontsize)
        self.ax.tick_params(axis='y', labelsize=self.fontsize)
        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)
        if self.x_formatter:
            self.ax.xaxis.set_major_formatter(self.x_formatter)
        if self.y_formatter:
            self.ax.yaxis.set_major_formatter(self.y_formatter)

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decorations()

        # Categorize by group and marker
        groups = self.data.groupby([self.group_name, "marker"])
        for name, group in groups:

            if len(name) > 1:
                name = name[0]

            if self.is_group_allowed(name):
                if self.marker != '':
                    mrk = self.marker
                else:
                    mrk = list(group["marker"])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                if self.xmolar:
                    xvals = val_el_to_mol(group[self.xvar])
                else:
                    xvals = group[self.xvar]

                if self.ymolar:
                    yvals = val_el_to_mol(group[self.yvar])
                else:
                    yvals = group[self.yvar]

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                sample_color = to_rgba(list(group["color"])[0], alpha=self.alpha_color)
                edge_color = to_rgba(list(group["color"])[0], alpha=self.alpha_edge_color)
                self.ax.scatter(
                    xvals, yvals,
                    edgecolors=edge_color,
                    marker=mrk, label=name,
                    facecolors=sample_color,
                    s=size,
                    zorder=zorder)

                if self.annotation:
                    for i, sample in xvals.iteritems():
                        self.ax.annotate(group[self.annotation].get(i), (xvals.get(i), yvals[i]), fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



import numpy as np
from georunes.plot.base import DiagramBase
from georunes.plot.helpers import LegendDrawer, ArrowDrawer
from georunes.tools.language import format_chemical_formula as _fml, get_translator


# Peccerillo, A. and Taylor, S.R., 1976. Geochemistry of Eocene calc-alkaline volcanic rocks from the Kastamonu area,
# northern Turkey. Contributions to mineralogy and petrology, 58(1), pp.63-81.

class DiagramPecceTaylor(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title=_fml("K2O vs SiO2") + "classification diagram (Peccerillo and Taylor 1976)",
                 padding={"bottom": 0.20}, annotation=None, **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, padding=padding, **kwargs)

        self.xlabel = _fml("SiO2 (wt%)")
        self.ylabel = _fml("K2O (wt%)")
        self.annotation = annotation

    def set_decoration(self):
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(45, 80)
        self.ax.set_ylim(0, 6)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        lx1 = np.array([48, 52, 56, 63, 70, 78])
        ly1 = np.array([0.3, 0.5, 0.7, 1, 1.3, 1.6])

        lx2 = np.array([48, 52, 56, 63, 70, 77])
        ly2 = np.array([1.2, 1.5, 1.8, 2.4, 3, 3.6])

        lx3 = np.array([48, 52, 56, 63, 70])
        ly3 = np.array([1.6, 2.4, 3.2, 4, 4.8])

        self.ax.plot(lx1, ly1, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx2, ly2, color=self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx3, ly3, color=self.decor_line_col, linewidth=0.6)

        txt_series = [
            [58, 0.2, _("low-K tholeite")],
            [53, 1, _("medium-K calc-alkaline")],
            [56, 2.6, _("high-K calc-alkaline")],
            [52, 3.6, _("shoshonitic")]
        ]
        for elt in txt_series:
            self.ax.text(*elt, fontsize="x-small")

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.exclude_groups and name not in self.exclude_groups:

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.ax.scatter(group["SiO2"], group["K2O"], edgecolors=group["color"],
                                marker=list(group["marker"])[0], label=name, facecolors=group["color"],
                                s=self.markersize,
                                alpha=0.9, zorder=zorder)

                if self.annotation:
                    for i, sample in group["SiO2"].iteritems():
                        self.ax.annotate(group[self.annotation].get(i), (group["SiO2"].get(i), group["K2O"].get(i)),
                                         fontsize='xx-small')

        self.plot_arrows()
        self.plot_legend()



from georunes.tools.filemanager import FileManager
from georunes.modmin.norm.cipw import CIPWNorm
from georunes.modmin.optim.bvls import BVLS
from georunes.modmin.optim.gd import GradientDescent
from georunes.modmin.optim.randsearch import RandomSearch
from georunes.modmin.optim.nnls import NNLS

filemanager = FileManager.get_instance()

#source = "D:/Dev/py/graph/geoglyph/georunes/modmin/cipw.csv"
source2 = "D:/Dev/py/graph/geoglyph/georunes/modmin/cipw2.csv"

#data = filemanager.read_file(source)

data2 = filemanager.read_file(source2)
minerals = filemanager.read_file("D:/Dev/py/graph/geoglyph/georunes/modmin/minerals_t.csv")

vb = 2

cipw = CIPWNorm(verbose=vb)
nnls = NNLS(verbose=vb)
bvls = BVLS(verbose=vb)
rs = RandomSearch(verbose=vb, dist_func="max")
gd = GradientDescent(verbose=3, dist_func="SMAPE")

# cipw.compute(data, skip_cols=1, minor_included=True,  co2_calcite=1, co2_cancrinite=0)


max_minerals = {}
min_minerals = {}
# max_minerals["Hematite"] = 0.12
min_minerals["Hematite"] = 0.04  # check if it is a full mineral, not element of solid solution
ratios = {"PG": [["Albite", "Anorthite"], [60, 40]],
          "Mica": [["Biotite", "Muscovite"], [8, 12]]}
ratios = {"PG": [["Albite", "Anorthite"], [50, 50]],
          "Mica": [["Biotite", "Muscovite"], [10, 10]]}

ratios=None
bvls.compute(data2, skip_cols=4, raw_minerals_data= minerals, ignore_oxides=("LOI",), ratios = ratios,
max_minerals = max_minerals, min_minerals=min_minerals, unfillable_partitions_allowed = False)
#
# nnls.compute(data2, skip_cols=4, raw_minerals_data= minerals, ignore_oxides=("LOI",), ratios= ratios)


starter = {
    "Quartz": 19,
    "Hematite": 10,
    "Orthoclase": 15,
    'PG': 38,
    'Mica': 18
}

# p,s = gd.compute(data2, skip_cols=5, raw_minerals_data= minerals, ignore_oxides=("LOI",),
#                  max_iter=100000, learn_rate=0.0001, ratios=ratios, max_minerals = max_minerals, min_minerals=min_minerals,
#                  starting_partition = starter
#                  )


# p,s = rs.compute(data2, skip_cols=4, raw_minerals_data= minerals, ignore_oxides=("LOI"), #starting_partition = starter,
#                  max_iter=1000, search_semiedge=0.0001, ratios=ratios, max_minerals = max_minerals, min_minerals=min_minerals,
#                  unfillable_partitions_allowed = False
#                  )








from georunes.plot.base import DiagramBase
from georunes.plot.helpers import ArrowDrawer, LegendDrawer
from georunes.tools.language import format_chemical_formula as _fml, get_translator


# Frost, B.R., Barnes, C.G., Collins, W.J., Arculus, R.J., Ellis, D.J. and Frost, C.D., 2001. A geochemical
# classification for granitic rocks. Journal of petrology, 42(11), pp.2033-2048.

class DiagramFrostSiFeNb(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title=_fml("SiO2 vs Fe# Frost diagram"), **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, **kwargs)
        if 'FeO' not in self.data.keys():
            raise Exception("FeO data missing. Check the table.")
        self.xlabel = _fml("SiO2 (wt %)")
        self.ylabel = "FeO/(FeO + MgO)"

    def set_decoration(self):
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(50, 80)
        self.ax.set_ylim(0.4, 1)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        x1 = (50, 80)
        y1 = (0.676, 0.814)

        self.ax.plot(x1, y1, self.decor_line_col, linewidth=0.6)

        self.ax.text(52, 0.8, _("ferroan"), size="small")
        self.ax.text(52, 0.6, _("magnesian"), size="small")

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):

                fenb = group["FeO"] / (group["FeO"] + group["MgO"])
                si = (group["SiO2"])

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.ax.scatter(si, fenb, edgecolors=group["color"],
                                marker=list(group["marker"])[0], label=name, facecolors=group["color"],
                                s=self.markersize,
                                alpha=0.7, zorder=zorder)

        self.plot_arrows()
        self.plot_legend()


class DiagramFrostSiFeTotNb(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title=_fml("SiO2 vs Fe* Frost diagram"), calc_total_Fe=False, **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, **kwargs)
        if 'FeOt' not in self.data.keys():
            if calc_total_Fe:
                if 'FeO' not in self.data.keys():
                    raise Exception("FeO data missing. Cannot calculate total FeO.")
                elif 'Fe2O3' not in self.data.keys():
                    raise Exception("Fe2O3 data missing. Cannot calculate total FeO.")
                else:
                    self.data['FeOt'] = self.data['FeO'] + self.data["Fe2O3"] * 0.8998
            else:
                raise Exception("FeOt data missing. Check the table.")
        self.xlabel = _fml("SiO2 (wt %)")
        self.ylabel = "FeO$^{tot}$/(FeO$^{tot}$ + MgO)"

    def set_decoration(self):
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(50, 80)
        self.ax.set_ylim(0.4, 1)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        x1 = (50, 80)
        y1 = (0.716, 0.854)

        self.ax.plot(x1, y1, self.decor_line_col, linewidth=0.6)

        self.ax.text(52, 0.8, _("ferroan"), size="small")
        self.ax.text(52, 0.6, _("magnesian"), size="small")

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):

                fenb = group["FeOt"] / (group["FeOt"] + group["MgO"])
                si = (group["SiO2"])

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.ax.scatter(si, fenb, edgecolors=group["color"],
                                marker=list(group["marker"])[0], label=name, facecolors=group["color"],
                                s=self.markersize,
                                alpha=0.7, zorder=zorder)

        self.plot_arrows()
        self.plot_legend()


class DiagramFrostSiMALI(DiagramBase, ArrowDrawer, LegendDrawer):
    def __init__(self, datasource, title=_fml("SiO2 vs modified alkali-lime index Frost diagram"), **kwargs):
        DiagramBase.__init__(self, datasource=datasource, title=title, **kwargs)
        self.xlabel = _fml("SiO2 (wt %)")
        self.ylabel = _fml("Na2O + K2O - CaO (wt %)")

    def set_decoration(self):
        if not self.no_title:
            self.ax.set_title(self.title, size=self.title_fs)
        _ = get_translator(self.lang_cfg)
        self.ax.set_xlim(50, 80)
        self.ax.set_ylim(-8, 12)

        self.ax.set_xlabel(self.xlabel, fontsize=self.fontsize)
        self.ax.set_ylabel(self.ylabel, fontsize=self.fontsize)

        lx = [i for i in range(50, 75, 5)]

        ly1 = [-41.86 + 1.112 * x - 0.00572 * x * x for x in lx]
        ly2 = [-44.72 + 1.094 * x - 0.00527 * x * x for x in lx]
        ly3 = [-45.36 + 1.0043 * x - 0.00427 * x * x for x in lx]

        self.ax.plot(lx, ly1, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx, ly2, self.decor_line_col, linewidth=0.6)
        self.ax.plot(lx, ly3, self.decor_line_col, linewidth=0.6)

        self.ax.text(51, 2.5, _("alkalic"), size="small", )
        self.ax.text(51, -1.3, _("alkali-calcic"), size="small", rotation=30)
        self.ax.text(51, -3.5, _("calc-alkalic"), size="small", rotation=30)
        self.ax.text(51, -6.5, _("calcic"), size="small", )

    def plot(self):
        DiagramBase.plot_config(self)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):

                mali = group["Na2O"] + group["K2O"] - group["CaO"]
                si = (group["SiO2"])

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.ax.scatter(si, mali, edgecolors=group["color"],
                                marker=list(group["marker"])[0], label=name, facecolors=group["color"],
                                s=self.markersize,
                                alpha=0.7, zorder=zorder)

        self.plot_arrows()
        self.plot_legend()



import itertools
import warnings
from pathlib import Path

import matplotlib.cm as cm
import pandas as pd
from matplotlib.colors import to_hex

from georunes.tools.data import unique

_vtc = ('color', 'marker', 'order', 'zorder', 'label',)

_markers = ("o", "s", "p", "*", "<", ">", "^", "8", "h", "H", "+",
            "x", "X", "D", "d", "v", "1", "2" "3", "4", "|", "_", ".", 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)


def check_data(dataset, group_name="group", supp_group=None, supp_vtc=None, ignore_checking_markers=False):
    if supp_group:
        list_group = [group_name, *supp_group]
    else:
        list_group = group_name

    if supp_vtc:
        vtc = itertools.chain(_vtc, supp_vtc)
    else:
        vtc = _vtc

    if ignore_checking_markers:
        a = list(vtc)
        a.remove('marker')
        vtc = tuple(a)

    if group_name not in dataset.keys():
        raise ValueError("Group name '" + str(group_name) + "' not found in dataset")

    groups = dataset.groupby(list_group)
    for name, group in groups:

        for field in vtc:
            if field in group.columns:

                test_value = list(group[field])[0]

                for val in group[field]:
                    if val != test_value:
                        raise ValueError("Different values for group " + str(name) +
                                         " and field " + str(field) + " : " +
                                         str(test_value) + " and " + str(val))


def populate_file(file, sheet=0, group_name='group', cmap='viridis',
                  pop_marker=True, pop_zorder=False, pop_label=False, output_suffix="_updated",
                  sep=",", delimiter=None):
    fname = Path(file).stem
    fext = Path(file).suffix
    parent_dir = Path(file).parent
    if fext in (".xls", ".xlsx") and sheet == 0:
        warnings.warn("No sheet name specified, the first sheet of the Excel file will be used.")

    # Load data
    if fext in (".xls", ".xlsx"):
        data = pd.read_excel(file, sheet_name=sheet)
    elif fext in (".csv", ".txt"):
        data = pd.read_csv(file, sep=sep, delimiter=delimiter)
    if group_name not in data.columns:
        raise ValueError("The specified 'group_name' parameter is not in the excel sheet.")

    unit_data = data[group_name].tolist()  # Column of categories
    units = unique(unit_data)  # All the categories to update in file
    # Colors data to add
    i_cmap = cm.get_cmap(cmap, len(units))
    rgbcolors = i_cmap(range(len(units)))
    colors = [to_hex(i) for i in rgbcolors]

    markers = _markers[0:len(units)]  # Markers data to add

    # Create dictionaries for data attribution
    dict_colors = dict(zip(units, colors[0:len(units)]))
    if pop_marker:
        dict_markers = dict(zip(units, markers))
    if pop_zorder:
        dict_zorders = dict(zip(units, range(4, 4 + len(units))))
    if pop_label:
        dict_labels = dict(zip(units, units))

    # Create the new contents of the columns to edit
    color_data, marker_data, zorder_data, label_data = [], [], [], []
    for unit in unit_data:
        color_data.append(dict_colors[unit])
        if pop_marker:
            marker_data.append(dict_markers[unit])
        if pop_zorder:
            zorder_data.append(dict_zorders[unit])
        if pop_label:
            label_data.append(dict_labels[unit])

    # Populate columns data
    data['color'] = color_data
    if pop_marker:
        data['marker'] = marker_data
    if pop_zorder:
        data['zorder'] = zorder_data
    if pop_label:
        data['label'] = label_data

    # Save new file
    new_fname = str(parent_dir) + '\\' + fname + output_suffix + fext
    if fext in (".xls", ".xlsx"):
        new_sheet = sheet if sheet != 0 else "Sheet1"
        data.to_excel(new_fname, sheet_name=new_sheet, index=False)
    elif fext in (".csv", ".txt"):
        data.to_csv(new_fname, sep=sep)
    print("Edited file saved as '" + new_fname + "'")


def populate_file_with_config(file, config, sheet=0, group_name='group', output_suffix="cfg_update",
                              sep=",", delimiter=None, ):
    fname = Path(file).stem
    fext = Path(file).suffix
    parent_dir = Path(file).parent
    if fext in (".xls", ".xlsx") and sheet == 0:
        warnings.warn("No sheet name specified, the first sheet of the Excel file will be used.")

    # Check configuration
    attribs = config.keys()
    if 'group' not in attribs:
        raise ValueError("The parameter 'group' is missing in the config dictionary.")

    units = config['group']  # All the categories to update in file
    if len(unique(units)) < len(units):
        raise ValueError("A doublon found in definition of group parameters in configuration.")

    card = len(config['group'])
    for key in attribs:
        if key not in ('group', 'label') and len(config[key]) != card:
            msg = "The length of the parameter '" + key + \
                  "' is different from the number of group defined in the config."
            raise ValueError(msg)

    # Load data
    if fext in (".xls", ".xlsx"):
        data = pd.read_excel(file, sheet_name=sheet)
    elif fext in (".csv", ".txt"):
        data = pd.read_csv(file, sep=sep, delimiter=delimiter)
    if group_name not in data.columns:
        raise ValueError("The specified 'group_name' parameter is not in the excel sheet.")

    unit_data = data[group_name].tolist()  # Column of categories

    # Data to add
    if not all([x in config['group'] for x in unique(unit_data)]):
        raise ValueError("Some groups existing in data file miss their configuration.")
    if 'color' in attribs:
        colors = config['color']
    if 'marker' in attribs:
        markers = config['marker']
    if 'zorder' in attribs:
        zorders = config['zorder']
    if 'label' in attribs:
        labels = config['label']

    # Create dictionaries for data attribution
    if 'color' in attribs:
        dict_colors = dict(zip(units, colors[0:len(units)]))
    if 'marker' in attribs:
        dict_markers = dict(zip(units, markers))
    if 'zorder' in attribs:
        dict_zorders = dict(zip(units, zorders))
    if 'labels' in attribs:
        if labels == 'auto' or labels is None:
            dict_labels = dict(zip(units, units))
        else:
            dict_labels = dict(zip(units, labels))

    # Create the new contents of the columns to edit
    color_data, marker_data, zorder_data, label_data = [], [], [], []
    for unit in unit_data:
        if 'color' in attribs:
            color_data.append(dict_colors[unit])
        if 'marker' in attribs:
            marker_data.append(dict_markers[unit])
        if 'zorder' in attribs:
            zorder_data.append(dict_zorders[unit])
        if 'labels' in attribs:
            label_data.append(dict_labels[unit])

    # Populate columns data
    if 'color' in attribs:
        data['color'] = color_data
    if 'marker' in attribs:
        data['marker'] = marker_data
    if 'zorder' in attribs:
        data['zorder'] = zorder_data
    if 'labels' in attribs:
        data['label'] = label_data

    # Save new file
    new_fname = str(parent_dir) + '\\' + fname + output_suffix + fext
    if fext in (".xls", ".xlsx"):
        new_sheet = sheet if sheet != 0 else "Sheet1"
        data.to_excel(new_fname, sheet_name=new_sheet, index=False)
    elif fext in (".csv", ".txt"):
        data.to_csv(new_fname, sep=sep)
    print("Edited file saved as '" + new_fname + "'")


def check_geochem_res(reservoir, listing=None):  # Listing for eventually required elements
    chklist = ['alias', 'color']
    if listing is not None:
        chklist = chklist.append(*listing)

    keys = reservoir.keys()
    for el in chklist:
        if el not in keys:
            raise ValueError("Key ", el, "not found in reservoir configuration.")

    return True



import numpy as np
from pandas import DataFrame
from scipy.optimize import nnls
from georunes.modmin.optim.base import Optimizer


class NNLS(Optimizer):
    def __init__(self, verbose=False, **kwargs):
        Optimizer.__init__(self, verbose=verbose, **kwargs)
        self.dist_func = "euclidian"
        self.notif = ">>>>>> Non-Negative Least Squares method"

    def _compute(self, raw_data, skip_cols, raw_minerals_data, to_round=4, ignore_oxides=None,
                 max_iter=None, residual_in_suppl=False):
        if self.verbose > 1:
            print("Round digits :", to_round, " --  Maximum iterations :", max_iter,
                  " --  Oxides to ignore : " + str(ignore_oxides) if ignore_oxides else "")
        self.prepare_data(raw_data, skip_cols, raw_minerals_data, ignore_oxides)

        partitions = DataFrame(columns=self.list_minerals)
        deviation_name = "deviation_" + self.dist_func
        suppl = DataFrame(columns=[deviation_name])
        bulk_chems = [0]*len(self.data.index)
        found_chems = [0]*len(self.data.index)

        for i in self.data.index:
            if self.verbose: print(">>> Composition", i)
            list_minerals_i = self.list_minerals.copy()
            minerals_data_i = self.minerals_data.copy()
            bulk_chems[i] = self.data.iloc[i].to_numpy()

            # Ignore minerals missing some oxides
            unnecessary_minerals = []
            for ox in self.list_bulk_ox:
                if self.data.loc[i, ox] == 0:
                    for mineral in minerals_data_i:
                        if minerals_data_i.loc[ox, mineral] > 0:
                            minerals_data_i = minerals_data_i.drop(columns=mineral)
                            list_minerals_i.remove(mineral)
                            unnecessary_minerals.append(mineral)
            if self.verbose and unnecessary_minerals:
                print("Unnecessary minerals :", *unnecessary_minerals)

            # Direct calculation of the result
            result = nnls(minerals_data_i, bulk_chems[i], max_iter)
            dict_res = {list_minerals_i[i]: result[0][i] for i in range(len(list_minerals_i))}

            idx_new_row = len(partitions)
            for miner, prop in dict_res.items():
                partitions.loc[idx_new_row, miner] = prop
            partitions = partitions.fillna(0)
            partitions.iloc[idx_new_row] = 100 * partitions.iloc[idx_new_row]
            partitions = partitions.round(to_round)

            deviation = self.deviation(bulk_chems[i], np.dot(self.minerals_data, partitions.iloc[idx_new_row] / 100))
            found_chems[i] = np.dot(self.minerals_data, partitions.loc[idx_new_row] / 100).round(to_round)
            suppl.loc[idx_new_row, deviation_name] = deviation
            suppl.loc[idx_new_row, "total_chem"] = sum(found_chems[i])

            if self.verbose:
                print("Solution", idx_new_row)
                print(partitions.loc[idx_new_row].to_dict())
                print("Corresponding composition")
                print([str(self.list_bulk_ox[p]) + " : " + str(found_chems[i]) for p in range(self.nb_oxides)])
                print("Deviation :", round(deviation, to_round), "%" if self.dist_func == "SMAPE" else "", "\n------")

        partitions["Total"] = partitions.sum(axis=1).round(to_round)
        suppl["diff_total_chem"] = suppl["total_chem"] - self.init_total
        if residual_in_suppl:
            residuals = np.subtract(bulk_chems, found_chems)
            t_residuals = list(map(list, zip(*residuals)))  # Transposition
            for p, ox in enumerate(self.list_bulk_ox):
                print(t_residuals[p])
                suppl["resid_" + str(ox)] = t_residuals[p]

        return partitions, suppl



import math
import numpy as np


def nan_to_none(x, y):
    nx, ny = x, y
    if x is not None and math.isnan(x):
        nx = None
    if y is not None and math.isnan(y):
        ny = None
    return nx, ny


def min_wo_none(x, y):
    x, y = nan_to_none(x, y)
    if (x is None) and (y is None):
        return np.nan
    elif x is None:
        return y
    elif y is None:
        return x
    else:
        return min(x, y)


def gg_max(x, y):
    x, y = nan_to_none(x, y)
    if (x is None) and (y is None):
        return np.nan
    elif x is None:
        return y
    elif y is None:
        return x
    else:
        return max(x, y)


# Function to get min in each colum of two rows
def row_min(old_row, cur_row):
    if not old_row:
        return cur_row

        # Check if same length
    if len(old_row) != len(cur_row):
        raise ValueError('Two rows doesn\' have the same size !')

    new_row = []
    for x, y in zip(old_row, cur_row):
        new_row.append(min_wo_none(x, y))
    return new_row


# Function to get max in each colum of two rows
def row_max(old_row, cur_row):
    if not old_row:
        return cur_row

    # Check if same length
    if len(old_row) != len(cur_row):
        raise ValueError('Two rows doesn\' have the same size !')

    new_row = []
    for x, y in zip(old_row, cur_row):
        new_row.append(gg_max(x, y))

    return new_row


def unique(list1):
    unique_list = []
    for x in list1:
        if x not in unique_list:
            unique_list.append(x)
    return unique_list



import numpy as np
from pandas import DataFrame
from georunes.modmin.optim.base import Optimizer, is_in_bounds, random_part_with_bounds, random_part_in_hypercube


class RandomSearch(Optimizer):

    def __init__(self, dist_func="euclidian", **kwargs):
        Optimizer.__init__(self, **kwargs)
        self.dist_func = dist_func
        self.notif = ">>>>>> Random research Method / deviation function : " + dist_func

    def _compute(self, raw_data, skip_cols, raw_minerals_data, to_round=4,
                 max_minerals=None, min_minerals=None, unfillable_partitions_allowed=True, ignore_oxides=None,
                 max_iter=1000000, limit_deviation=1e-03, search_semiedge=1, scale_semiedge=None,
                 starting_partition=None, force_totals=False, residual_in_suppl=False):

        if self.verbose > 1:
            print("Round digits :", to_round, " --  Maximum iterations :", max_iter, " --  Limit deviation:",
                  limit_deviation,
                  " --  Semiegde of hypercube research :", search_semiedge,
                  " --  Oxides to ignore : " + str(ignore_oxides) if ignore_oxides else "")
            if starting_partition: print("Starting partition :", starting_partition)

        self.prepare_data(raw_data, skip_cols, raw_minerals_data, ignore_oxides)
        partitions = DataFrame(columns=self.list_minerals)
        deviation_name = "deviation_" + self.dist_func
        suppl = DataFrame(columns=[deviation_name])
        bulk_chems = [0] * len(self.data.index)
        found_chems = [0] * len(self.data.index)

        if force_totals:
            if unfillable_partitions_allowed:
                raise Exception("The parameter force_totals is True. The parameter unfillable_partitions_allowed "
                                "should be set to False. Check the computing configuration.")
            target_totals = self.init_total
        else:
            target_totals = [100] * len(self.data.index)

        if not scale_semiedge or scale_semiedge < 0 or scale_semiedge > 1:
            if self.verbose > 0:
                print("WARNING : The scale must be superior to 0 and inferior or equals to 1. Value set to 1.")
            scale_semiedge = 1

        # Get maximum proportion for each mineral
        for i in self.data.index:
            if self.verbose: print(">>> Composition", i)
            list_minerals_i = self.list_minerals.copy()
            minerals_data_i = self.minerals_data.copy()
            bulk_chems[i] = self.data.iloc[i].to_numpy()
            search_semiedge_i = search_semiedge

            # Get maximum possible proportion for each mineral
            max_minerals_prop = []
            unnecessary_minerals = []
            for mineral in self.list_minerals:
                # Maximum proportion for each oxide equals Ox wt% in bulk chemistry / ox wt% in mineral composition
                max_oxides_prop = self.data.iloc[i] / minerals_data_i[mineral].transpose()
                mineral_prop = max_oxides_prop.min()
                if np.isinf(mineral_prop) or mineral_prop == 0:
                    minerals_data_i = minerals_data_i.drop(columns=mineral)
                    list_minerals_i.remove(mineral)
                    unnecessary_minerals.append(mineral)
                else:
                    max_minerals_prop.append(mineral_prop)
            nb_minerals_i = len(list_minerals_i)
            min_minerals_prop = [0] * nb_minerals_i

            # Update minerals proportions with defined settings if provided
            if max_minerals:
                for key, value in max_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        max_minerals_prop[idx] = min(max_minerals_prop[idx], max_minerals[key])
            if min_minerals:
                for key, value in min_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        min_minerals_prop[idx] = max(min_minerals_prop[idx], min_minerals[key])
            if not unfillable_partitions_allowed and sum(max_minerals_prop) < 1.05:  # 5% tolerance
                raise Exception("Problem in composition " + str(i)
                                + ". The sum of the maximum proportions of minerals cannot complete to 100 "
                                  "(found " + str(100 * sum(max_minerals_prop)) + str(")."))

            if self.verbose and unnecessary_minerals:
                print("Unnecessary minerals :", *unnecessary_minerals)

            # Random calculation
            if self.verbose: print("Start calculations. Expected number of iterations :", max_iter)
            min_deviation = 1e10
            # Starting value
            if isinstance(starting_partition, dict):
                if not unfillable_partitions_allowed and sum(starting_partition.values()) != 100:
                    print("WARNING : The starting partition does not complete to 100. The calculations will be "
                          "started with a random composition.")
                    candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                        verbose=self.verbose)
                else:
                    # Check if all minerals in the config are present in minerals_data_i
                    if all(el in list_minerals_i for el in starting_partition.keys()):
                        candidate = []
                        for key in list_minerals_i:
                            val = starting_partition[key] / 100 if key in starting_partition.keys() else 0
                            candidate.append(val)
                        if not is_in_bounds(candidate, max_minerals_prop, min_minerals_prop):
                            print("WARNING : The starting partition does not respect the provided and calculated "
                                  "bounds data : Minerals ,", list_minerals_i, "Max ,", max_minerals_prop, "Min ,",
                                  min_minerals_prop, "The calculations will be started with a random composition.")
                            candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                                unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                                verbose=self.verbose)
                    else:
                        print("WARNING : Some minerals in starting partition are not present in mineral chemistry "
                              "data. The calculations will be started with a random composition.")
                        candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                            unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                            verbose=self.verbose)
            else:
                candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                    unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                    verbose=self.verbose)

            if self.verbose: print("Starting partition", dict(zip(list_minerals_i, candidate)))

            # Loop
            for k in range(max_iter):
                if search_semiedge_i > 1:
                    raise Exception("The variable search_semiedge_i must be inferior or equal to 1.")
                elif search_semiedge_i < 1 and candidate:
                    new_candidate = random_part_in_hypercube(candidate, search_semiedge_i, max_minerals_prop,
                                                             min_minerals_prop, total=target_totals[i] / 100,
                                                             unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                             verbose=self.verbose)
                    if set(new_candidate) == set(candidate) and search_semiedge_i < 1:
                        search_semiedge_i = search_semiedge_i * scale_semiedge
                        if search_semiedge_i < pow(10, -to_round):
                            if self.verbose: print(
                                "Updated semiedge is inferior to the rounding precision. Search ended")
                            break
                        if self.verbose > 0:
                            print("In iteration", k, ", search_semiedge_i changed to", search_semiedge_i)

                else:
                    new_candidate = random_part_with_bounds(nb_minerals_i, max_minerals_prop, min_minerals_prop,
                                                            total=target_totals[i] / 100,
                                                            unfillable_partitions_allowed=unfillable_partitions_allowed,
                                                            verbose=self.verbose)

                corresp_chem = np.dot(minerals_data_i, new_candidate).round(decimals=to_round)
                dist = self.deviation(bulk_chems[i], corresp_chem)
                if dist < min_deviation:
                    min_deviation = dist
                    candidate = new_candidate
                    if min_deviation < limit_deviation:
                        if self.verbose: print("Bottom deviation condition reached after iteration", k)
                        break
                    if self.verbose > 1:
                        print("Better solution at iteration", k)
                        print(dict(zip(list_minerals_i, candidate)))
                        print("New deviation :", min_deviation, "%" if self.dist_func == "SMAPE" else "")

                if self.verbose and k == max_iter - 1:
                    print("Max iterations reached")

            dict_res = {list_minerals_i[i]: round(candidate[i], to_round) for i in range(len(list_minerals_i))}

            idx_new_row = len(partitions)
            for miner, prop in dict_res.items():
                partitions.loc[idx_new_row, miner] = prop
            partitions = partitions.fillna(0)
            partitions.iloc[idx_new_row] = 100 * partitions.iloc[idx_new_row]
            partitions = partitions.round(to_round)

            deviation = self.deviation(bulk_chems[i], np.dot(self.minerals_data, partitions.iloc[idx_new_row] / 100))
            found_chems[i] = np.dot(self.minerals_data, partitions.loc[idx_new_row] / 100).round(to_round).tolist()
            suppl.loc[idx_new_row, deviation_name] = deviation
            suppl.loc[idx_new_row, "total_chem"] = sum(found_chems[i])
            if self.verbose:
                print("Solution", idx_new_row)
                print(partitions.loc[idx_new_row].to_dict())
                print("Corresponding composition")
                print([str(self.list_bulk_ox[p]) + " : " + str(found_chems[i]) for p in range(self.nb_oxides)])
                print("Deviation :", round(deviation, to_round), "%" if self.dist_func == "SMAPE" else "", "\n------")

        partitions["Total"] = partitions.sum(axis=1).round(to_round)
        suppl["diff_total_chem"] = suppl["total_chem"] - self.init_total
        if residual_in_suppl:
            residuals = np.subtract(bulk_chems, found_chems)
            t_residuals = list(map(list, zip(*residuals)))  # Transposition
            for p, ox in enumerate(self.list_bulk_ox):
                print(t_residuals[p])
                suppl["resid_" + str(ox)] = t_residuals[p]

        return partitions, suppl



import numpy as np
from pandas import DataFrame
from scipy.optimize import lsq_linear
from georunes.modmin.optim.base import Optimizer


class BVLS(Optimizer):
    def __init__(self, **kwargs):
        Optimizer.__init__(self, **kwargs)
        self.dist_func = "euclidian"
        self.notif = ">>>>>> Bounded Variable Least Squares Method"

    def _compute(self, raw_data, skip_cols, raw_minerals_data, to_round=4, max_minerals=None, min_minerals=None,
                 unfillable_partitions_allowed=True, ignore_oxides=None, residual_in_suppl=False):
        if self.verbose > 1:
            print("Round digits :", to_round, " --  Oxides to ignore : " + str(ignore_oxides) if ignore_oxides else "")
        self.prepare_data(raw_data, skip_cols, raw_minerals_data, ignore_oxides)
        partitions = DataFrame(columns=self.list_minerals)
        deviation_name = "deviation_" + self.dist_func
        suppl = DataFrame(columns=[deviation_name])
        bulk_chems = [0] * len(self.data.index)
        found_chems = [0] * len(self.data.index)

        for i in self.data.index:
            if self.verbose: print(">>> Composition", i)
            list_minerals_i = self.list_minerals.copy()
            minerals_data_i = self.minerals_data.copy()
            bulk_chems[i] = self.data.iloc[i].to_numpy()

            # Get maximum possible proportion for each mineral
            max_minerals_prop = []
            unnecessary_minerals = []
            for mineral in self.list_minerals:
                # Maximum proportion for each oxide equals Ox wt% in bulk chemistry / Ox wt% in mineral composition
                max_oxides_prop = self.data.iloc[i] / minerals_data_i[mineral].transpose()
                mineral_prop = max_oxides_prop.min()
                if np.isinf(mineral_prop) or mineral_prop == 0:
                    minerals_data_i = minerals_data_i.drop(columns=mineral)
                    list_minerals_i.remove(mineral)
                    unnecessary_minerals.append(mineral)
                else:
                    max_minerals_prop.append(mineral_prop)
            nb_minerals_i = len(list_minerals_i)
            min_minerals_prop = [0] * nb_minerals_i

            # Update minerals proportions with defined settings if provided
            if max_minerals:
                for key, value in max_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        max_minerals_prop[idx] = min(max_minerals_prop[idx], max_minerals[key])
            if min_minerals:
                for key, value in min_minerals.items():
                    idx = list_minerals_i.index(key)
                    if idx:
                        min_minerals_prop[idx] = max(min_minerals_prop[idx], min_minerals[key])
            if not unfillable_partitions_allowed and sum(max_minerals_prop) < 1.05:  # 5% tolerance
                raise Exception("Problem in composition " + str(i)
                                + ". The sum of the maximum proportions of minerals cannot complete to 100 "
                                  "(found " + str(100 * sum(max_minerals_prop)) + str(")."))

            if self.verbose and unnecessary_minerals:
                print("Unnecessary minerals :", *unnecessary_minerals)

            # Direct calculation of the result
            result = lsq_linear(minerals_data_i, bulk_chems[i], (min_minerals_prop, max_minerals_prop), method='bvls',
                                verbose=min(self.verbose, 2))
            dict_res = {list_minerals_i[i]: result.x[i] for i in range(len(list_minerals_i))}

            idx_new_row = len(partitions)
            for miner, prop in dict_res.items():
                partitions.loc[idx_new_row, miner] = prop
            partitions = partitions.fillna(0)
            partitions.iloc[idx_new_row] = 100 * partitions.iloc[idx_new_row]
            partitions = partitions.round(to_round)

            deviation = self.deviation(bulk_chems[i], np.dot(self.minerals_data, partitions.iloc[idx_new_row] / 100))
            found_chems[i] = np.dot(self.minerals_data, partitions.loc[idx_new_row] / 100).round(to_round)
            suppl.loc[idx_new_row, deviation_name] = deviation
            suppl.loc[idx_new_row, "total_chem"] = sum(found_chems[i])

            if self.verbose:
                print("Solution", idx_new_row)
                print(partitions.loc[idx_new_row].to_dict())
                print("Corresponding composition")
                print([str(self.list_bulk_ox[p]) + " : " + str(found_chems[i]) for p in range(self.nb_oxides)])
                print("Deviation :", round(deviation, to_round), "%" if self.dist_func == "SMAPE" else "", "\n------")

        partitions["Total"] = partitions.sum(axis=1).round(to_round)
        suppl["diff_total_chem"] = suppl["total_chem"] - self.init_total
        if residual_in_suppl:
            residuals = np.subtract(bulk_chems, found_chems)
            t_residuals = list(map(list, zip(*residuals)))  # Transposition
            for p, ox in enumerate(self.list_bulk_ox):
                print(t_residuals[p])
                suppl["resid_" + str(ox)] = t_residuals[p]

        return partitions, suppl



import warnings
from pathlib import Path
import numpy as np
import pandas as pd


# Singleton model, from github.com/pazdera/1098129
class FileManager:
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if FileManager.__instance is None:
            FileManager()
        return FileManager.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if FileManager.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            FileManager.__instance = self

        self.files = []
        self.datas = dict()

    def read_file(self, datasource, sheet_name=None, sep=",", delimiter=None, index=None):
        fext = Path(datasource).suffix

        if fext in (".xls", ".xlsx") and sheet_name == 0:
            warnings.warn("No sheet name specified, the first sheet of the Excel file will be used.")

        ref_end = '_' + str(sheet_name) if sheet_name and sheet_name != 0 else ''
        ref = str(datasource) + ref_end
        if ref in self.files:  # Check if already loaded
            return self.datas[ref]
        else:
            if fext in (".xls", ".xlsx"):
                data = pd.read_excel(datasource, sheet_name=sheet_name, index_col=index)
            elif fext in (".csv", ".txt"):
                data = pd.read_csv(datasource, sep=sep, delimiter=delimiter, index_col=index)
            else:
                msg = "Extension file " + fext + " not recognized."
                raise Exception(msg)

            if not data.empty:
                self.files.append(ref)

                # Preprocess
                data = data.replace('bdl', np.nan)

                self.datas[ref] = data
                print(ref + " data loaded.")
                return data



from setuptools import setup, find_packages

setup(
    packages=find_packages(include=['georunes*'], ),
    python_requires=">=3.6",
    install_requires=["matplotlib",
                      "numpy",
                      "scipy",
                      "pandas",
                      "python-ternary"],
    include_package_data=True,
    package_data={"georunes": ["data/*.txt", "data/*.csv"]}
)



from georunes.plot.helpers import ArrowDrawerTernary
from georunes.plot.ternary.ternbase import DiagramTernaryBase
from georunes.tools.chemistry import val_ox_to_el, val_ox_to_mc
from georunes.tools.colors import darken_color
from georunes.tools.language import get_translator
from georunes.tools.plotting import normalize_marker_size


# Jensen, L.S., 1976. A new plot for classifying subalkalic volcanic rocks. Ontario Division of Mines,
# Miscellaneous Paper, 66, pp.1-22.

class DiagramJensen(DiagramTernaryBase, ArrowDrawerTernary):
    def __init__(self, datasource, title="Ternary plot of Jensen (1976)", **kwargs):
        DiagramTernaryBase.__init__(self, datasource=datasource, title=title,
                                    top_label='Fet + Ti', left_label='Al', right_label='Mg',
                                    **kwargs)

    def set_decoration(self, ):
        # Order for points : right var, top var, left var
        _ = get_translator(self.lang_cfg)

        p10 = (0, 10, 90)
        p11 = (100 - 29 - 53, 29, 53)
        p111 = (100 - 29.5 - 52, 29.5, 52)
        p12 = (100 - 29 - 51, 29, 51)
        p13 = (100 - 26 - 51, 26, 51)
        p14 = (100 - 12 - 51.5, 12, 51.5)

        p31 = (5, 15, 80)
        p32 = (20, 0, 80)

        p41 = (0, 30, 70)
        p42 = (30, 0, 70)

        p51 = (0, 40, 60)
        p52 = (30, 10, 60)

        p61 = (67 / 2, 33, 67 / 2)
        p62 = (100 - 25 - 51, 25, 51)

        p71 = (0, 50, 50)
        p72 = (17, 100 - 17 - 50, 50)
        p73 = (20, 29, 51)
        curve7 = [p71, p72, p73]

        p81 = (60, 40, 0)
        p82 = (60, 0, 40)

        p91 = (22, 56, 22)
        p92 = (50, 0, 50)

        curve = [p10, p11, p111, p12, p13, p14]

        self.tax.plot(curve, linewidth=1.0, color=self.decor_line_col)
        self.tax.plot(curve7, linewidth=1.0, color=self.decor_line_col, linestyle="--")
        self.tax.line(p31, p32, linewidth=1, color=self.decor_line_col, linestyle="--")
        self.tax.line(p41, p42, linewidth=1, color=self.decor_line_col, linestyle="--")
        self.tax.line(p51, p52, linewidth=1, color=self.decor_line_col, linestyle="--")
        self.tax.line(p61, p62, linewidth=1, color=self.decor_line_col, linestyle="--")
        self.tax.line(p81, p82, linewidth=1, color=self.decor_line_col, linestyle="--")
        self.tax.line(p91, p92, linewidth=1, color=self.decor_line_col)
        self.tax.annotate(_("TH"), (2, 45,), fontsize="small", color=darken_color(self.decor_text_col), rotation=30,
                          fontweight='bold')
        self.tax.annotate(_("Dacite"), (4, 25,), color=self.decor_text_col, fontsize="x-small", rotation=-60)
        self.tax.annotate(_("High-Fe\ntholeiite\nbasalt"), (16, 44,), ha="center", color=self.decor_text_col,
                          fontsize="x-small")
        self.tax.annotate(_("Andesite"), (9, 31,), color=self.decor_text_col, fontsize="x-small", rotation=-60)
        self.tax.annotate(_("Rhyolite"), (1, 21,), color=self.decor_text_col, fontsize="x-small", )

        self.tax.annotate(_("CA"), (40, 3), fontsize="small", color=darken_color(self.decor_text_col), rotation=30,
                          fontweight='bold')
        self.tax.annotate(_("Andesite"), (21, 7,), color=self.decor_text_col, fontsize="x-small", rotation=-60)
        self.tax.annotate(_("Dacite"), (15, 3,), color=self.decor_text_col, fontsize="x-small", rotation=-60)
        self.tax.annotate(_("Basalt"), (27, 12,), color=self.decor_text_col, fontsize="x-small", rotation=-60)
        self.tax.annotate(_("High-Mg\ntholeiite\nbasalt"), (40, 10,), ha="center", color=self.decor_text_col,
                          fontsize="x-small")
        self.tax.annotate(_("Komatiitic basalt"), (35, 35,), color=self.decor_text_col, fontsize="x-small")
        self.tax.annotate(_("Komatiite"), (70, 15,), color=self.decor_text_col, fontsize="x-small")

    def plot(self):
        DiagramTernaryBase.plot_config(self, no_gridline=True)
        self.set_decoration()

        # Categorize by group
        groups = self.data.groupby(self.group_name)
        for name, group in groups:

            if self.is_group_allowed(name):
                top_var = val_ox_to_mc(group['Fe2O3t']) + val_ox_to_el(group['TiO2'])
                left_var = val_ox_to_mc(group['Al2O3'])
                right_var = val_ox_to_mc(group['MgO'])
                var_sum = top_var + left_var + right_var
                norm_z = 100 * left_var / var_sum
                norm_x = 100 * right_var / var_sum
                norm_y = 100 * top_var / var_sum

                if self.marker != '':
                    mrk = self.marker
                else:
                    mrk = list(group["marker"])[0]

                if self.marker_size_scaled():
                    size = normalize_marker_size(group[self.markersize['var_scale']], self.markersize['val_max'],
                                                 self.markersize['val_min'], self.markersize['size_max'],
                                                 self.markersize['size_min'])
                else:
                    size = self.markersize

                # Convert points to tuples
                points = [(norm_x.get(i), norm_y.get(i), norm_z.get(i)) for i, sample in top_var.iteritems()]

                zorder = 4
                if self.drawing_order:
                    zorder = list(group[self.drawing_order])[0]

                self.tax.scatter(points, edgecolors=group["color"],
                                 marker=mrk, label=name, facecolors=group["color"], s=size,
                                 alpha=self.alpha_color, zorder=zorder)

                if self.annotation:
                    for i, sample in top_var.iteritems():
                        self.tax.annotate(group[self.annotation].get(i), (norm_x.get(i), norm_y.get(i)),
                                          fontsize='xx-small')

        self.plot_legend()
